from enum import Enum
from dataclasses import dataclass
from abc import abstractmethod, ABC


class BoundaryType(Enum):
    SLIP_FREE_ENDS = "slip_free_ends"
    CLAMPED = "clamped"
    CONTACT = "contact"


class PressBoundary(ABC):
    """Dataclass for press boundary condition parameters."""
    bc_type: BoundaryType

    def __init__(self, bc_type: BoundaryType, **kwargs):
        self.bc_type = bc_type

    @classmethod
    def create(cls, bc_type: BoundaryType, **kwargs) -> 'PressBoundary':
        if bc_type == BoundaryType.CONTACT:
            return PressBoundaryContact(**kwargs)
        elif bc_type == BoundaryType.CLAMPED:
            return PressBoundaryClamped(**kwargs)
        elif bc_type == BoundaryType.SLIP_FREE_ENDS:
            return PressBoundarySlipFreeEnds(**kwargs)
        else:
            raise ValueError(f"Unknown boundary condition type: {bc_type}")

    @property
    @abstractmethod
    def snes_options(self) -> str:
        pass

    @abstractmethod
    def options(self, center, radius, height, load_fraction) -> str:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass


class PressBoundaryContact(PressBoundary):
    friction_coefficient: float

    def __init__(self, friction_coefficient: float = 0.5, **kwargs):
        super().__init__(bc_type=BoundaryType.CONTACT, **kwargs)
        self.friction_coefficient = friction_coefficient

    @property
    def snes_options(self) -> str:
        if not hasattr(self, '_snes_options'):
            self._snes_options = '\n'.join([
                "# SNES options for contact boundary conditions",
                "snes:",
                "  monitor:",
                "  max_it: 15",
                "  rtol: 1e-6",
                "augmented_lagrangian_inner_snes:",
                "  linesearch:",
                "    type: bt",
                "    monitor:",
                "  max_it: 20",
                "  monitor:",
                "  ksp:",
                "    ew:",
                "    ew_version: 3",
                "    ew_rtol0: 1e-4",
                "    ew_rtolmax: 1e-4",
                ""
            ])
        return self._snes_options

    def options(self, center, radius, height, load_fraction) -> str:
        if self.friction_coefficient == 0.0:
            friction_options = "\n".join([
                "    friction:",
                "      type: none",
            ])
        else:
            friction_options = '\n'.join([
                "    friction:",
                "      type: coulomb",
                f"      kinetic: {self.friction_coefficient}",
                "      penalty_min: 10",
                "      penalty_max: 5e3",
                "      penalty_scale: 2",
            ])
        penalty_ops = '\n'.join([
            "    penalty_min: 100",
            "    penalty_max: 5e4",
            "    penalty_scale: 4",
        ])
        ops = '\n'.join([
            "bc:",
            "  allow_no_clamp:",
            "  contact: 1,2,3,4,5,6",
            "  # Bottom",
            "  contact_1:",
            "    shape: platen",
            "    normal: 0,0,1",
            f"    center: {center[0]},{center[1]},{center[2]}",
            penalty_ops,
            "    type: augmented_lagrangian",
            friction_options,
            "  # Top, compressing 40%",
            "  contact_2:",
            "    shape: platen",
            penalty_ops,
            "    normal: 0,0,-1",
            f"    center: {center[0]},{center[1]},{center[2]+height}",
            f"    distance: {load_fraction * height} # load_fraction * height",
            "    type: augmented_lagrangian",
            friction_options,
            "  # Outside",
            "",
        ])
        for i in range(3, 7):
            ops += '\n'.join([
                f"  contact_{i}:",
                "    shape: cylinder",
                "    axis: 0,0,1",
                f"    radius: {radius}",
                f"    center: {center[0]},{center[1]},{center[2]}",
                "    inside:",
                penalty_ops,
                "    type: augmented_lagrangian",
                friction_options,
                "",
            ])
        return ops

    @property
    def name(self) -> str:
        friction_str = f"mu{self.friction_coefficient}" if self.friction_coefficient != 0.0 else "frictionless"
        return f"{self.bc_type.value}_{friction_str}"

    def __str__(self) -> str:
        friction_str = f"Coulomb friction with μ={self.friction_coefficient}" if self.friction_coefficient != 0.0 else "frictionless"
        return f"Contact Boundaries, {friction_str}"


class PressBoundaryClamped(PressBoundary):
    def __init__(self, **kwargs):
        super().__init__(bc_type=BoundaryType.CLAMPED, **kwargs)

    @property
    def snes_options(self) -> str:
        if not hasattr(self, '_snes_options'):
            self._snes_options = '\n'.join([
                "# SNES options for slip/clamped boundary conditions",
                "snes:",
                "  linesearch:",
                "    type: bisection",
                "    monitor:",
                "  max_it: 20",
                "  monitor:",
                "  ksp:",
                "    ew:",
                "    ew_version: 3",
                "    ew_rtol0: 1e-4",
                "    ew_rtolmax: 1e-4",
                "",
            ])
        return self._snes_options

    def options(self, center, radius, height, load_fraction) -> str:
        return '\n'.join([
            "bc:",
            "  clamp: 1,2",
            "  # Clamped displacement for top and bottom",
            "  clamp_2:",
            f"    translate: 0,0,{-load_fraction * height} # -load_fraction * height",
            "  # Prevent x,y expansion beyond the die boundary",
            "  slip: 3,4,5,6",
            "  slip_3:",
            "    components: 0,1",
            "  slip_4:",
            "    components: 0,1",
            "  slip_5:",
            "    components: 0,1",
            "  slip_6:",
            "    components: 0,1",
            "",
        ])

    @property
    def name(self) -> str:
        return f"{self.bc_type.value}"

    def __str__(self) -> str:
        return "Clamped Boundaries"


class PressBoundarySlipFreeEnds(PressBoundary):
    def __init__(self, **kwargs):
        super().__init__(bc_type=BoundaryType.SLIP_FREE_ENDS, **kwargs)

    @property
    def snes_options(self) -> str:
        if not hasattr(self, '_snes_options'):
            self._snes_options = '\n'.join([
                "# SNES options for slip/clamped boundary conditions",
                "snes:",
                "  linesearch:",
                "    type: bisection",
                "    monitor:",
                "  max_it: 20",
                "  monitor:",
                "  ksp:",
                "    ew:",
                "    ew_version: 3",
                "    ew_rtol0: 1e-4",
                "    ew_rtolmax: 1e-4",
                "",
            ])
        return self._snes_options

    def options(self, center, radius, height, load_fraction) -> str:
        return '\n'.join([
            "bc:",
            "  slip: 1,2,3,4,5,6",
            "  # Allow x,y displacement for top and bottom, prescribe z displacement",
            "  slip_1:",
            "    components: 2",
            "  slip_2:",
            "    components: 2",
            f"    translate: {-load_fraction * height} # -load_fraction * height",
            "  slip_3:",
            "    components: 0,1",
            "  slip_4:",
            "    components: 0,1",
            "  slip_5:",
            "    components: 0,1",
            "  slip_6:",
            "    components: 0,1",
            "",
        ])

    @property
    def name(self) -> str:
        return f"{self.bc_type.value}"

    def __str__(self) -> str:
        return "Clamped Wall Boundary with Free-Slip Boundaries on Top and Bottom"



import typer
import logging
import rich
from rich.logging import RichHandler
from rich.theme import Theme
from pathlib import Path
from typing import Optional

from .helper import build, config

try:
    from .mpm.main import app as mpm_app
    HAVE_MPM = True
except ImportError:
    HAVE_MPM = False


try:
    from .postprocess.main import app as postprocess_app
    HAVE_POSTPROCESS = True
except ImportError:
    HAVE_POSTPROCESS = False

custom_theme = Theme({
    "info": "dim white",
    "warning": "bold yellow",
    "warn": "bold yellow",
    "error": "bold red",
    "err": "bold red",
    "success": "bold green",
    "h1": "bold underline green",
    "h2": "bold underline white",
})

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.WARNING, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
rich.reconfigure(theme=custom_theme, soft_wrap=True)

app = typer.Typer()


# @app.callback()
def main(
    ratel_dir: Optional[Path] = None,
    output_dir: Optional[Path] = None,
    scratch_dir: Optional[Path] = None,
):
    """
    Ratel Runner
    """
    if ratel_dir is not None:
        config.set('RATEL_DIR', f"{ratel_dir.resolve()}")
    if output_dir is not None:
        config.set('OUTPUT_DIR', f"{output_dir.resolve()}")
    if scratch_dir is not None:
        config.set('SCRATCH_DIR', f"{scratch_dir.resolve()}")


# Basic helper utilities
app.add_typer(config.app, name="config")
app.add_typer(build.app, name="build")

if HAVE_MPM:
    app.add_typer(mpm_app, name="mpm", help="Run iMPM Experiments")

if HAVE_POSTPROCESS:
    app.add_typer(postprocess_app, name="postprocess", help="Post-process iMPM Experiments")
    # # Press experiments
    # press_app = typer.Typer()
    # mpm_app.add_typer(press_app, name="press", help="Press consolidation experiments")
    # press_app.add_typer(press_sticky_air.app)
    # press_app.add_typer(press_no_air.app)
    # mpm_app.add_typer(sweep.app, name="sweep")
    # perf_app = typer.Typer()
    # # Performance experiments
    # mpm_app.add_typer(perf_app, name="performance", help="Performance experiments")
    # perf_app.add_typer(efficiency.app)

if __name__ == "__main__":
    app()



from platform import machine
import typer
import json
from pathlib import Path
from rich import print
import os
from typing import Annotated, Any, Callable
from dataclasses import dataclass
from difflib import get_close_matches
from rich.table import Table
from enum import Enum
from copy import deepcopy
from contextlib import contextmanager

from .flux import machines

__all__ = ['get_app_dir', 'get', 'set', 'unset', 'get_fallback', 'app']


class FileOpenMode(Enum):
    READ = 'r'
    WRITE = 'w'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class GPUMode(Enum):
    SPX = 'SPX'
    CPX = 'CPX'
    TPX = 'TPX'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


@dataclass
class ConfigKey:
    name: str
    description: str
    type: type
    check: Callable[[Any], bool] = lambda _: True

    def __eq__(self, key):
        if isinstance(key, ConfigKey):
            return self.name == key.name
        return self.name == key

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class CheckBounded():
    def __init__(self, lower=None, upper=None, inclusive_lower=False, inclusive_upper=False):
        assert lower is not None or upper is not None, "One of upper or lower bounds must be set"
        self.lower = lower
        self.upper = upper
        self.inclusive_lower = inclusive_lower
        self.inclusive_upper = inclusive_upper

    def __str__(self) -> str:
        if self.lower is None:
            return f"must be <{('=' if self.inclusive_upper else '')} {self.upper}"
        if self.upper is None:
            return f"must be >{('=' if self.inclusive_lower else '')} {self.lower}"
        bracket_lower = "[" if self.inclusive_lower else "("
        bracket_upper = "]" if self.inclusive_upper else ")"
        return f"must be in the interval {bracket_lower}{self.lower}, {self.upper}{bracket_upper}"

    def __call__(self, value) -> bool:
        good = value is not None
        if good and self.lower is not None:
            good = good and (value > self.lower or (self.inclusive_lower and value == self.lower))
        if good and self.upper is not None:
            good = good and (value < self.upper or (self.inclusive_upper and value == self.upper))
        if not good:
            print(f'[error]value {value} {self}')
            return False
        return True


_KNOWN_KEYS = {
    "PETSC_DIR": ConfigKey("PETSC_DIR", "PETSc repository directory", Path),
    "PETSC_ARCH": ConfigKey("PETSC_ARCH", "PETSc arch name", Path),
    "PETSC_CONFIG": ConfigKey("PETSC_CONFIG", "User-specified configuration file for PETSc", Path),
    "LIBCEED_DIR": ConfigKey("LIBCEED_DIR", "libCEED repository directory", Path),
    "RATEL_DIR": ConfigKey("RATEL_DIR", "Ratel repository directory", Path),
    "SCRATCH_DIR": ConfigKey("SCRATCH_DIR", "Directory in which to write simulation outputs", Path),
    "OUTPUT_DIR": ConfigKey("OUTPUT_DIR", "Directory in which to place symlinks to output directories", Path),
    "GPU_MODE": ConfigKey("GPU_MODE", "GPU mode, only matters for Tuolumne", GPUMode),
}


def add_key(key_name: str, key: ConfigKey):
    if key_name in _KNOWN_KEYS.keys():
        print(f'[warn]{key_name} already registered, overwriting')
    _KNOWN_KEYS[key_name] = key


def _get_config(machine: machines.Machine | None = None, name: str | None = None):
    """
    Get the configuration file, creating it if it does not exist.
    """
    if machine is None:
        machine = machines.detect_machine()
    app_dir = get_app_dir()
    if name is not None:
        config_file = Path(name)
        if not config_file.is_absolute():
            config_file = app_dir / config_file
        if not config_file.exists():
            config_file.parent.mkdir(parents=True, exist_ok=True)
            config = {}
            config_file.write_text(json.dumps(config, indent=2))
        return config_file
    if machine is None or machine == machines.Machine.DEFAULT:
        config_file = app_dir / "config.json"
    else:
        config_file = app_dir / f'{machine.value.lower()}' / "config.json"
    if not config_file.exists():
        config_file.parent.mkdir(parents=True, exist_ok=True)
        config = {}
        config_file.write_text(json.dumps(config, indent=2))
    return config_file


@contextmanager
def runtime_config(mode: FileOpenMode, machine: machines.Machine | None = None):
    """
    Read/write values in the application configuration file.
    """
    if machine is None:
        machine = machines.detect_machine()
    config_path = _get_config(machine)
    configuration = json.loads(config_path.read_text())
    yield configuration
    if mode == FileOpenMode.WRITE:
        config_path.write_text(json.dumps(configuration, indent=2))
    del configuration


app = typer.Typer(help="Manage Ratel Runner runtime configuration")


def get_app_dir():
    return Path(typer.get_app_dir("ratel-runner"))


def unset(key: str, machine: machines.Machine | None = None):
    """
    Remove a key from the runtime configuration.
    """
    with runtime_config(FileOpenMode.WRITE, machine=machine) as config:
        if key in config:
            del config[key]


@app.command('unset')
def unset_cmd(key: str, machine: machines.Machine | None = None):
    """
    Remove a key from the configuration file.
    """
    unset(key, machine=machine)


def set(key: str, value: str, machine: machines.Machine | None = None, quiet: bool = True):
    """
    Set a key-value pair in the runtime configuration.
    """
    if key not in _KNOWN_KEYS.keys():
        print(f"[warn]Unknown key:[/warn] {key}")
        similar = get_close_matches(key, _KNOWN_KEYS.keys())
        if len(similar):
            print(f"[warn]  Similar keys:[/warn] {' '.join(similar)}")
        raise typer.Exit(1)
    config_key = _KNOWN_KEYS[key]
    if not config_key.check(config_key.type(value)):
        raise typer.Abort(f'Error: Invalid value {value} for key {key}')
    with runtime_config(FileOpenMode.WRITE, machine=machine) as config:
        if not quiet:
            if config_key.name in config:
                print(f"Key {key} already exists. Overwriting value (old: {config[key]})")
            print(f"Setting {key} to {value}")
        config[config_key.name] = str(config_key.type(value))


@app.command('set')
def set_cmd(key: str, value: str, machine: machines.Machine | None = None):
    """
    Set a key-value pair in the configuration file.
    """
    set(key, value, machine=machine, quiet=False)


def get(key: str, machine: machines.Machine | None = None, quiet: bool = True):
    """
    Internal helper function to get the value of a key from the config file.
    """
    if key not in _KNOWN_KEYS.keys():
        if not quiet:
            print(f"[warn]Unknown key:[/warn] {key}")
        similar = get_close_matches(key, _KNOWN_KEYS.keys())
        if len(similar) and not quiet:
            print(f"[warn]  Similar keys:[/warn] {' '.join(similar)}")
        raise typer.Exit(1)
    config_key = _KNOWN_KEYS[key]
    with runtime_config(FileOpenMode.READ, machine=machine) as config:
        value = config.get(key, None)
    return config_key.type(value) if value else None


@app.command('get')
def get_cmd(key: str, machine: machines.Machine | None = None,
            script: Annotated[bool, typer.Option("-s", help="Output only the value")] = False):
    """
    Get the value of a key in the configuration file.
    """
    val = get(key, machine=machine, quiet=script)
    if script:
        if val is None:
            raise typer.Exit(1)
        print(val)
    else:
        print(f"{key}: ", "Key not found." if val is None else val)


def print_config(config: dict):
    """
    Print the configuration in a human-readable format.
    """
    if len(config) == 0:
        print("[success]Configuration empty, use[/success] `config set` [success]to add configuration variables")
    else:
        table = Table("Name", "Description", "Options", "Value")
        for key, value in config.items():
            config_key = _KNOWN_KEYS.get(key, ConfigKey(f'[warn]{key}', "[warn]UNKNOWN KEY[/warn]", str))
            options = ", ".join(str(e) for e in config_key.type) if issubclass(config_key.type, Enum) else ""
            table.add_row(f'[bold underline]{config_key.name}', f"{config_key.description}", options, value)
        print(table)


def list_config(machine: machines.Machine | None = None):
    """
    List all keys in the configuration file.
    """
    with runtime_config(FileOpenMode.READ, machine=machine) as config:
        print_config(config)


@app.command('list')
def list_cmd(machine: machines.Machine | None = None):
    """
    List all keys and values in the configuration file.
    """
    list_config(machine=machine)


@app.command('copy')
def copy_cmd(src: machines.Machine, dst: Annotated[machines.Machine | None, typer.Argument()] = None):
    """
    Copy all configuration variables from one machine to another
    """
    with runtime_config(FileOpenMode.READ, machine=src) as src_config:
        with runtime_config(FileOpenMode.WRITE, machine=dst) as dst_config:
            dst_config.clear()
            for key, value in src_config.items():
                dst_config[key] = value


@app.command('dump')
def dump_cmd(dst: Path, machine: machines.Machine | None = None):
    """
    Dump the configuration file as JSON
    """
    full_dst = dst.resolve()
    full_dst.parent.mkdir(parents=True, exist_ok=True)
    with runtime_config(FileOpenMode.READ, machine=machine) as config:
        full_dst.write_text(json.dumps(config, indent=2))
    print(f'[success]Wrote configuration to {full_dst}[/success]')


@app.command('load')
def load_cmd(src: Path, machine: machines.Machine | None = None):
    """
    Load the configuration file from JSON
    """
    full_src = src.resolve()
    if not full_src.exists():
        print(f'[error]Source file {full_src} does not exist[/error]')
        raise typer.Exit(1)
    config_data = json.loads(full_src.read_text())
    with runtime_config(FileOpenMode.WRITE, machine=machine) as config:
        config.clear()
        for key, value in config_data.items():
            config[key] = value
    print(f'[success]Loaded configuration from {full_src}[/success]')


def get_fallback(key: str, default=None, machine: machines.Machine | None = None):
    """
    Get the value of a key from the runtime config, environment variable, default provided, or raise an error.
    """
    value = get(key, machine=machine)
    if value is not None:
        return value
    value = os.environ.get(key, None)
    if value is not None:
        return value
    if default is not None:
        return default
    else:
        raise ValueError(
            f"{key} not set. Please set the {key} environment variable or use the config command to set it.")


stash_app = typer.Typer()
app.add_typer(stash_app, name='stash', help='Stash and apply runtime configurations')


@contextmanager
def stash(mode: FileOpenMode = FileOpenMode.WRITE, machine: machines.Machine | None = None):
    """
    Context manager for stashed configurations
    """
    app_dir = get_app_dir()
    if machine is None or machine == machines.Machine.DEFAULT:
        stash_path = app_dir / "stashes.json"
    else:
        stash_path = app_dir / f'{machine.value.lower()}' / "stashes.json"
    stash_path = stash_path
    stash_data = json.loads(stash_path.read_text())
    yield stash_data
    if mode == FileOpenMode.WRITE:
        stash_path.write_text(json.dumps(stash_data, indent=2))
    del stash_data


@stash_app.command('push')
def stash_push(name: Annotated[str | None, typer.Argument()] = None,
               machine: machines.Machine | None = None, quiet: bool = False):
    """
    Stash the current configuration under a given name
    """
    with stash(mode=FileOpenMode.WRITE, machine=machine) as stash_data:
        with runtime_config(FileOpenMode.READ, machine=machine) as config:
            if name is None:
                name = f'stash-{len(stash_data["stashes"])+1}'
            stash_data['stashes'][name] = deepcopy(config)
            stash_data['stack'].append(name)
            if not quiet:
                print(f'Stashed current configuration as [bold]{name}[/bold]')


@stash_app.command('apply')
def stash_apply(name: Annotated[str | None, typer.Argument()] = None,
                machine: machines.Machine | None = None, quiet: bool = False) -> str:
    """
    Apply the stashed configuration with the given name
    """
    with stash(mode=FileOpenMode.READ, machine=machine) as stash_data:
        with runtime_config(FileOpenMode.WRITE, machine=machine) as config:
            if name is None:
                if len(stash_data['stack']) == 0:
                    if not quiet:
                        print('[error]No stashed configurations to apply[/error]')
                    raise typer.Exit(1)
                resolved_name: str = stash_data['stack'][-1]
            else:
                resolved_name = name
            if resolved_name not in stash_data['stashes']:
                print(f'[error]No stashed configuration with name {resolved_name}[/error]')
                raise typer.Exit(1)
            stashed_config = stash_data['stashes'][resolved_name]
            config.clear()
            for key, value in stashed_config.items():
                config[key] = value
            if not quiet:
                print(f'Applied stashed configuration [bold]{resolved_name}[/bold]:')
                print_config(config)
    return resolved_name


@stash_app.command('remove')
def stash_remove(name: str, machine: machines.Machine | None = None):
    """
    Remove the stashed configuration with the given name without applying it
    """
    with stash(mode=FileOpenMode.WRITE, machine=machine) as stash_data:
        if name not in stash_data['stashes']:
            print(f'[error]No stashed configuration with name {name}[/error]')
            raise typer.Exit(1)
        stash_data['stack'].remove(name)
        del stash_data['stashes'][name]


@stash_app.command('pop')
def stash_pop(name: Annotated[str | None, typer.Argument()] = None,
              machine: machines.Machine | None = None, quiet: bool = False):
    """
    Apply the stashed configuration with the given name and remove it from the stash
    """
    name = stash_apply(name=name, machine=machine, quiet=quiet)
    stash_remove(name=name, machine=machine)


@stash_app.command('show')
def stash_show(name: str, machine: machines.Machine | None = None, quiet: bool = False):
    """
    Show the stashed configuration with the given name without applying it
    """
    with stash(mode=FileOpenMode.READ, machine=machine) as stash_data:
        if name not in stash_data['stashes']:
            print(f'[error]No stashed configuration with name {name}[/error]')
            raise typer.Exit(1)
        stashed_config = stash_data['stashes'][name]
        if not quiet:
            print(f'Stashed configuration [bold]{name}[/bold]:')
            print_config(stashed_config)


@stash_app.command('peek')
def stash_peek(machine: machines.Machine | None = None):
    """
    Show the most recently stashed configuration without applying it
    """
    with stash(mode=FileOpenMode.READ, machine=machine) as stash_data:
        if len(stash_data['stack']) == 0:
            print('[error]No stashed configurations to show[/error]')
            raise typer.Exit(1)
        name = stash_data['stack'][-1]
        stashed_config = stash_data['stashes'][name]
        print(f'Stashed configuration [bold]{name}[/bold]:')
        print_config(stashed_config)


@stash_app.command('list')
def stash_list(machine: machines.Machine | None = None):
    """
    List all stashed configurations
    """
    with stash(mode=FileOpenMode.READ, machine=machine) as stash_data:
        if len(stash_data['stashes']) == 0:
            print('[success]No stashed configurations[/success]')
            return
        table = Table("Index", "Name")
        for order, name in enumerate(stash_data['stack'][::-1], start=1):
            table.add_row(str(order), f'[bold]{name}[/bold]')
        print(table)


@stash_app.command('clear')
def stash_clear(machine: machines.Machine | None = None, quiet: bool = False):
    """
    Clear all stashed configurations
    """
    with stash(mode=FileOpenMode.WRITE, machine=machine) as stash_data:
        stash_data['stack'] = []
        stash_data['stashes'] = {}
    if not quiet:
        print('[success]Cleared all stashed configurations[/success]')



import numpy as np
from matplotlib import pyplot as plt
from pathlib import Path
import pyvista as pv
from rich import print
import typer

import yaml

from orix.vector import Miller, Vector3d
from orix.quaternion import Orientation, Misorientation, symmetry
from diffpy.structure import Lattice, Structure
from orix.crystal_map import Phase
from orix import plot, data

app = typer.Typer()

# We'll want our plots to look a bit larger than the default size
plt.rcParams.update(
    {
        "figure.figsize": (10, 5),
        "lines.markersize": 2,
        "font.size": 15,
        "axes.grid": False,
    }
)
w, h = plt.rcParams["figure.figsize"]

monoclinic = Phase(
    point_group=symmetry.C2h,
    structure=Structure(lattice=Lattice(6.53, 11.03, 7.35, 90, 102.689, 90)),
)

triclinic = Phase(
    point_group=symmetry.C1,
    structure=Structure(lattice=Lattice(9.015, 9.0325, 6.825, 108.65, 91.98, 119.94)),
)


def vec_rotate(n, theta, v):
    n_dot_v = np.einsum("ij,ij->i", n, v)
    cos_theta = np.cos(theta)
    n_cross_v = np.cross(n, v)
    v_rot = cos_theta[:, None] * v + (np.sin(theta))[:, None] * n_cross_v + (n_dot_v * (1 - cos_theta))[:, None] * n
    return v_rot


def plot_orientations(orientation_axes, orientation_angles, orientation_axes0,
                      orientation_angles0, phase, out="orientations.png"):
    O0 = Orientation.from_axes_angles(Vector3d(orientation_axes0), orientation_angles0, symmetry=phase.point_group)
    mO = Misorientation.from_axes_angles(Vector3d(orientation_axes), orientation_angles)
    O = mO * O0
    g = Miller(hkl=[0, 0, 1], phase=phase)
    g = g.symmetrise(unique=True)
    poles0 = O0.inv().outer(g, lazy=True, progressbar=True, chunk_size=2000)  # type: ignore
    poles = O.inv().outer(g, lazy=True, progressbar=True, chunk_size=2000)  # type: ignore
    alpha = 10000 / orientation_axes.shape[0]

    fig = plt.figure(figsize=(2 * h, 2 * h))
    subplot_kw = {"projection": "stereographic"}

    ax0 = fig.add_subplot(221, **subplot_kw)
    ax0.scatter(poles0, alpha=alpha)  # type: ignore
    ax0.set_title("Initial")

    ax1 = fig.add_subplot(222, **subplot_kw)
    ax1.pole_density_function(poles0)  # type: ignore
    ax1.set_title("Initial")

    ax2 = fig.add_subplot(223, **subplot_kw)
    ax2.scatter(poles, alpha=alpha)  # type: ignore
    ax2.set_title("Deformed")

    ax3 = fig.add_subplot(224, **subplot_kw)
    ax3.pole_density_function(poles)  # type: ignore
    ax3.set_title("Deformed")
    plt.savefig(out)


def read_mesh(file, time_step=0):
    """Read a mesh from a file."""
    reader: pv.XdmfReader = pv.get_reader(file)  # type: ignore
    reader.set_active_time_point(time_step)
    mesh: pv.DataSet = reader.read()
    grains = mesh.threshold(1.5, scalars='material', method='upper', preference='point')
    del mesh
    return grains


def compute_poles(mesh: pv.DataSet):
    props = np.asarray(mesh.point_data["elastic parameters"])
    state = np.asarray(mesh.point_data["model state"])
    npts = props.shape[0]
    print(f"Computing pole orientations for {npts} points.")
    F = state.reshape((npts, 3, 3)) + np.eye(3)
    Finv = np.linalg.inv(F)
    n0 = props[:, 21:24]
    theta0 = props[:, 24]
    V = np.zeros_like(n0)
    V[:, 2] = 1.
    ab_normal0 = vec_rotate(n0, theta0, V)
    # n_tilde = F^-T * n0
    ab_normal = np.einsum("ijk,ij->ik", Finv, ab_normal0)
    ab_normal /= np.linalg.norm(ab_normal, axis=1, keepdims=True)
    orientation_axes = np.cross(ab_normal0, ab_normal)
    axes = orientation_axes / np.linalg.norm(orientation_axes, axis=1, keepdims=True)
    angles = np.arccos(np.einsum("ij,ij->i", ab_normal, ab_normal0))
    axes0 = n0
    angles0 = theta0
    return axes, angles, axes0, angles0


@app.command()
def pole_diagram(
    run_dir: Path,
    swarm_file: str = "swarm.xdmf",
    time_step: int = 0,
    out: str = "orientations.png",
):
    """Plot pole diagrams from swarm data."""
    swarm_file_path = run_dir / swarm_file
    if not swarm_file_path.exists():
        if swarm_file_path.with_suffix(".xmf").exists():
            swarm_file_path.write_text(swarm_file_path.with_suffix(".xmf").read_text())
        else:
            raise FileNotFoundError(f"Swarm file {swarm_file_path} not found.")
    options_file = run_dir / "options.yml"
    with open(options_file, 'r') as file:
        data = yaml.full_load(file)
    try:
        grain_options = data['mpm']['grains']
        a = grain_options['a'] * 1e7  # convert to angstroms
        b = grain_options['b'] * 1e7  # convert to angstroms
        c = grain_options['c'] * 1e7  # convert to angstroms
        alpha = grain_options['alpha']
        beta = grain_options['beta']
        gamma = grain_options['gamma']
        lattice = Lattice(a, b, c, alpha, beta, gamma)
        if alpha == 90.0 and gamma == 90.0 and beta != 90.0:
            phase = Phase(
                point_group=symmetry.C2h,
                structure=Structure(lattice=lattice),
            )
            print(f"Using monoclinic phase with lattice parameters: {a}, {b}, {c}, {alpha}, {beta}, {gamma}")
        elif alpha != 90.0 and beta != 90.0 and gamma != 90.0:
            phase = Phase(
                point_group=symmetry.C1,
                structure=Structure(lattice=lattice),
            )
            print(f"Using triclinic phase with lattice parameters: {a}, {b}, {c}, {alpha}, {beta}, {gamma}")
        else:
            raise ValueError("Unsupported lattice parameters for pole figure plotting.")
    except KeyError | ValueError:
        # guess
        if "monoclinic" in f'{run_dir}':
            phase = monoclinic
            print(f"Using monoclinic phase with lattice parameters: {phase.structure.lattice}")
        else:
            phase = triclinic
            print(f"Using triclinic phase with lattice parameters: {phase.structure.lattice}")

    axes, angles, axes0, angles0 = compute_poles(read_mesh(run_dir / swarm_file, time_step=time_step))  # type: ignore
    print(f"Computed pole orientations for {axes.shape[0]} points.")
    plot_orientations(axes, angles, axes0, angles0, phase, out=out)
    print(f"Saved pole figure to {out}.")


if __name__ == "__main__":
    app()



import typer
import logging
import rich
from rich.logging import RichHandler
from rich.theme import Theme

from .plot.pole_diagram import app as plot_pole_diagram_app


custom_theme = Theme({
    "info": "dim white",
    "warning": "bold yellow",
    "warn": "bold yellow",
    "error": "bold red",
    "err": "bold red",
    "success": "bold green",
    "h1": "bold underline green",
    "h2": "bold underline white",
})

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.WARNING, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
rich.reconfigure(theme=custom_theme, soft_wrap=True)

app = typer.Typer()

plot_app = typer.Typer()
app.add_typer(plot_app, name="plot", help="Post-process plotting tools")
plot_app.add_typer(plot_pole_diagram_app)

if __name__ == "__main__":
    app()
